import os

import bpy

bl_info = {
    "name": "Super Texture",
    "category": "Node",
    "author": "Tim Crellin (Thatimster)",
    "blender": (4, 0, 2),
    "location": "Node Editor Toolbar",
    "description": "Generates multiple texture maps from a single image input",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "",
    "version": (1, 85)
}


class REMOVE_OT_supertexture(bpy.types.Operator):
    bl_idname = "remove.supertexture"
    bl_description = "Remove Super-Texture from source node"
    bl_label = "Remove"

    @classmethod
    def poll(cls, context):
        result = False
        if context.object and context.object.type == "MESH":
            has_mat = context.object.data.materials
            if len(has_mat) > 0:
                active_mat = has_mat[context.object.active_material_index]
                if active_mat and active_mat.node_tree:
                    current_nodes = active_mat.node_tree.nodes
                    for i in current_nodes:
                        if i.type == 'GROUP' and i.select and i.node_tree.name.startswith("Super-Texture"):
                            result = True
                            break
        return result

    def execute(self, context):
        active_mat = context.object.data.materials[context.object.active_material_index]
        to_remove = []
        ref = None
        last_active = None
        for current in active_mat.node_tree.nodes:
            if current.select and current.type == "GROUP" and current.node_tree.name.startswith("Super-Texture"):
                to_remove.append(current)

        for remove_group in to_remove:
            for node in remove_group.node_tree.nodes:
                if node.label == "REF":
                    ref = node
                    break
            if ref == None:
                ref = next(
                    n for n in remove_group.node_tree.nodes if n.type == "TEX_IMAGE")

            n_node = copy_node(active_mat.node_tree, ref, 1)[0]
            if n_node.type == "GROUP":
                n_node.name = n_node.node_tree.name

            if "Coordinates" in remove_group.inputs and remove_group.inputs["Coordinates"].links:
                prev = remove_group.inputs["Coordinates"].links[0]
                active_mat.node_tree.links.new(
                    prev.from_node.outputs[prev.from_socket.name], n_node.inputs[0])
            if remove_group.outputs[0].links:
                next_node = remove_group.outputs[0].links[0]
                active_mat.node_tree.links.new(
                    next_node.to_node.inputs[next_node.to_socket.name], n_node.outputs[0])

            n_node.location = remove_group.location
            active_mat.node_tree.nodes.remove(remove_group)
            last_active = n_node
        active_mat.node_tree.nodes.active = last_active
        for grp in bpy.data.node_groups:
            if not grp.users:
                bpy.data.node_groups.remove(grp)
        return {"FINISHED"}


class ADD_OT_supertexture(bpy.types.Operator):
    bl_idname = 'add.supertexture'
    bl_description = 'Apply Super Texture'
    bl_category = 'Node'
    bl_label = 'Apply'

    @classmethod
    def poll(cls, context):
        result = False
        if context.object and context.object.type == "MESH":
            has_mat = context.object.data.materials
            if len(has_mat) > 0:
                current = context.active_node
                if current:
                    if not current.name.startswith("Super-Texture"):
                        if current.type.startswith("TEX_") or current.type == "GROUP":
                            if current.select and current.type != "TEX_COORD":
                                if current.type == "TEX_IMAGE":
                                    if current.image:
                                        result = True
                                elif current.type == "GROUP":
                                    if current.node_tree and not current.node_tree.name.startswith("Super-Texture"):
                                        result = True
                                else:
                                    result = True
        return result

    def execute(self, context):
        active_mat = context.object.data.materials[context.object.active_material_index]
        if context.active_node.type != "TEX_IMAGE":
            context.scene.st_parallax = False
        new_node = self.append_setup(context)

        selected = context.active_node

        new_name = ""
        if selected.type == "TEX_IMAGE" and selected.image != None:
            new_name += "_" + selected.image.name
        elif selected.type == "GROUP":
            new_name += "_" + selected.node_tree.name
        else:
            new_name += "_" + selected.type[4:]
        to_explore = [(new_node, active_mat.node_tree)]
        explored = []
        to_modify = set()
        img_width = None
        img_height = None

        while to_explore:
            new, tr = to_explore.pop(0)
            if new.type == "GROUP" and new.node_tree.name not in explored:
                ns = new.node_tree.nodes
                to_explore += zip(ns, [new.node_tree] * len(ns))
                explored.append(new.node_tree.name)

            elif new.type == "TEX_IMAGE":
                to_modify.add((new, tr))
            elif new.type == "VALUE":
                if new.label == "img_width":
                    img_width = (new, tr)
                elif new.label == "img_height":
                    img_height = (new, tr)

        if selected.type == "TEX_IMAGE":
            for tm, _ in to_modify:
                tm.image = selected.image
                tm.projection = selected.projection
                tm.projection_blend = selected.projection_blend
                tm.interpolation = selected.interpolation
                tm.extension = selected.extension

            img_width[0].outputs[0].default_value = selected.image.size[0]
            img_height[0].outputs[0].default_value = selected.image.size[1]
        else:
            vect = [i for i in selected.inputs if i.type == "VECTOR"]
            col = [o for o in selected.outputs if o.type == "RGBA"]
            if vect and col and selected.type == 'GROUP':
                self.replace_nodes(to_modify, selected)
            else:
                # remove and replace norm-gen with bump node, add relevant contrast / gamma / rgb-bw
                nnt = new_node.node_tree
                nnt.links.new(nnt.nodes["Bump"].outputs[0],
                              nnt.nodes["Group Output"].inputs["Normal"])
                self.replace_nodes(to_modify, selected)
            # clean up and remove unused nodes
            for to_remove, tr in to_modify:
                tr.nodes.remove(to_remove)

        new_node.location = selected.location
        new_node.node_tree.name = "Super-Texture_" + new_name

        if selected.outputs[0].links:
            nt = active_mat.node_tree
            for l in selected.outputs[0].links:
                nt.links.new(new_node.outputs[0],
                             l.to_node.inputs[l.to_socket.name])
            dest = new_node.outputs[0].links[0].to_node
            if "Roughness" in dest.inputs:
                nt.links.new(
                    new_node.outputs["Roughness"], dest.inputs["Roughness"])
            if "Normal" in dest.inputs:
                nt.links.new(new_node.outputs["Normal"], dest.inputs["Normal"])
            if "Alpha" in dest.inputs:
                nt.links.new(new_node.outputs["Alpha"], dest.inputs["Alpha"])

        # sort out uv mapping
        tex_coords = [
            n for n in active_mat.node_tree.nodes if n.type == "TEX_COORD"]
        tex_coord = None
        if tex_coords:
            tex_coord = tex_coords[0]
        else:
            tex_coord = active_mat.node_tree.nodes.new(
                type="ShaderNodeTexCoord")
            tex_coord.location = (new_node.location.x-200, new_node.location.y)
        if selected.type != "TEX_IMAGE" and selected.type != "GROUP":
            active_mat.node_tree.links.new(
                tex_coord.outputs[0], new_node.inputs["Coordinates"])
        else:
            if (selected.inputs["Vector"].links):
                active_mat.node_tree.links.new(
                    selected.inputs["Vector"].links[0].from_node.outputs[selected.inputs["Vector"].links[0].from_socket.name], new_node.inputs["Coordinates"]
                )
            else:
                active_mat.node_tree.links.new(
                    tex_coord.outputs[2], new_node.inputs["Coordinates"]
                )
        active_mat.node_tree.nodes.remove(selected)

        # correct alpha
        if not len(new_node.node_tree.nodes["Group Output"].inputs["Alpha"].links):
            new_node.node_tree.nodes["Group Output"].inputs["Alpha"].default_value = 1

        self.set_default_values(context, new_node)
        return {"FINISHED"}

    def set_default_values(self, context, node):
        pref = context.preferences.addons[__package__].preferences

        for k, v in prop_ref.items():
            if (v in node.inputs):
                node.inputs[v].default_value = getattr(pref, k)

    def replace_nodes(self, to_modify, selected):
        for new, tr in to_modify:
            generated = copy_node(tr, selected, 1)[0]
            if new.label == "REF":
                generated.label = "REF"
            generated.name = new.name
            generated.location = new.location
            generated.hide = True
            input_l = new.inputs[0].links[0]
            tr.links.new(
                input_l.from_node.outputs[input_l.from_socket.name], generated.inputs[0])
            for l in new.outputs[0].links:
                tr.links.new(generated.outputs[0],
                             l.to_node.inputs[l.to_socket.name])

    def append_setup(self, context):
        mat_ref = {'st nor': ('Material.001exp', [
            'Super-Texture']), 'st par': ('Material.001exp', ['Super-Texture-Parallax'])}
        select_mat_name = 'st nor'
        if context.scene.st_parallax:
            select_mat_name = 'st par'
        select_mat = mat_ref[select_mat_name][0]
        file_loc = os.path.join(os.path.dirname(
            __file__), "data", "data.blend")
        if context.space_data.tree_type == "ShaderNodeTree" and context.space_data.shader_type == "OBJECT":
            with bpy.data.libraries.load(file_loc) as (data_from, data_to):
                data_to.materials = [
                    data_from.materials[data_from.materials.index(select_mat)]]
            target_tree = bpy.data.materials[select_mat]
            current_tree = context.object.active_material.node_tree
            target_nodes = [
                mn for mn in target_tree.node_tree.nodes if mn.name in mat_ref[select_mat_name][1]]
            current_nodes = current_tree.nodes
            current_group_nodes = [
                gn for gn in current_nodes if gn.type == "GROUP"]
            node_dict = {}
            for t in target_nodes:
                new_n = current_nodes.new(type=t.bl_idname)
                node_dict[t.name] = new_n

                if t.type == "GROUP":
                    if not t.node_tree.name.lower().startswith("super-texture"):
                        collect = sorted(
                            [n for n in current_group_nodes if t.label == n.name])
                        if collect:
                            new_n.node_tree = collect[0].node_tree
                    else:
                        new_n.width = 200
                        new_n.node_tree = t.node_tree
            for l in target_tree.node_tree.links:
                frm_n = l.from_node
                to_n = l.to_node
                if frm_n in target_nodes and to_n in target_nodes:
                    frm_sock_idx = list(frm_n.outputs).index(l.from_socket)
                    to_sock_idx = list(to_n.inputs).index(l.to_socket)
                    current_tree.links.new(
                        node_dict[frm_n.name].outputs[frm_sock_idx], node_dict[to_n.name].inputs[to_sock_idx])
            bpy.data.materials.remove(target_tree)
            return node_dict[list(node_dict.keys())[0]]


def copy_node(group, to_copy, num):
    # default node properties
    propstr = 'type,location,width,width_hidden,height,dimensions,name,label,inputs,outputs,internal_links,parent,use_custom_color,color,select,show_options,show_preview,hide,mute,show_texture,shading_compatibility'
    container = []

    for _ in range(num):
        container.append(group.nodes.new(type=to_copy.bl_idname))
    for n in container:
        mirror = [i for i in dir(n) if check_prop(propstr.split(","), i) and not i.startswith(
            "__") and not callable(getattr(n, i)) and not n.is_property_readonly(i)]
        for m in mirror:
            setattr(n, m, getattr(to_copy, m))
        n.name = to_copy.name

    # node inputs
    for val in range(len(to_copy.inputs)):
        for n_ in container:
            n_.inputs[val].default_value = to_copy.inputs[val].default_value
    return container


def check_prop(props, toCheck):
    if toCheck in props or toCheck.startswith("bl_"):
        return False
    return True


class SUPERTEXTURE_PT_panel(bpy.types.Panel):
    bl_space_type = 'NODE_EDITOR'
    bl_region_type = 'UI'
    bl_category = 'Texture-Tools'
    bl_label = 'Super Texture'

    @classmethod
    def poll(cls, context):
        if (context.space_data.tree_type == "ShaderNodeTree" and context.space_data.shader_type == "OBJECT"):
            return True
        else:
            return False

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        if context.scene.render.engine not in ["CYCLES", "BLENDER_EEVEE"]:
            row.label(text="仅 Cycles/Eevee 支持")
        else:
            if context.space_data.tree_type == "ShaderNodeTree" and context.space_data.shader_type == "OBJECT":
                if not context.object:
                    row.label(text="选择一个物体")
                elif not context.object.active_material:
                    row.label(text="添加一个材质")
                elif not context.object.active_material.node_tree:
                    row.label(text="启用使用节点")
                else:
                    if context.active_node and context.active_node.type == "TEX_IMAGE":
                        row.prop(context.scene, 'st_parallax')
                    col = layout.column(align=True)
                    row = col.row(align=True)
                    row.scale_y = 1.5
                    row.operator("add.supertexture", icon="RENDERLAYERS")
                    row.operator("remove.supertexture", icon="X")


prop_ref = {
    "ao": "AO",
    "height": "高度",
    "specular": "高光",
    "shadows": "移除阴影",
    "rough": "糙度",
    "rough_int": "糙度强度",
    "normal_size": "法线大小",
    "normal": "法线强度",
    "height": "高度",
    "height_contrast": "高度对比度",
    "height_blur": "高度模糊",
    "parallax_smooth": "视差平滑",
    "parallax": "视差偏移"
}


class SUPERTEXTURE_PR_preferences(bpy.types.AddonPreferences):
    bl_idname = __name__

    shadows: bpy.props.FloatProperty(
        default=0.25, name=prop_ref["shadows"], subtype="FACTOR", min=0, max=1)
    ao: bpy.props.FloatProperty(
        default=0.5, name=prop_ref["ao"], subtype="FACTOR", min=0, max=1)
    specular: bpy.props.FloatProperty(
        default=1, name=prop_ref["specular"], min=0, max=100)
    rough: bpy.props.FloatProperty(
        default=1, name=prop_ref["rough"], subtype="FACTOR", min=0, max=1)
    rough_int: bpy.props.FloatProperty(
        default=1.5, name=prop_ref["rough_int"])
    normal_size: bpy.props.FloatProperty(
        default=2, name=prop_ref["normal_size"])
    normal: bpy.props.FloatProperty(
        default=2, name=prop_ref["normal"], subtype="FACTOR", min=0, max=5)
    height: bpy.props.FloatProperty(
        default=1, name=prop_ref["height"])
    height_contrast: bpy.props.FloatProperty(
        default=0, name=prop_ref["height_contrast"])
    height_blur: bpy.props.FloatProperty(
        default=0, name=prop_ref["height_blur"])
    parallax_smooth: bpy.props.FloatProperty(
        default=0.2, name=prop_ref["parallax_smooth"], subtype="FACTOR", min=0, max=1)
    parallax: bpy.props.FloatProperty(
        default=0.15, name=prop_ref["parallax"], subtype="FACTOR", min=-1, max=1)

    def draw(self, context):
        layout = self.layout
        layout.label(text="默认值")
        row1 = layout.row()
        col = row1.column()
        for p in list(prop_ref.keys())[:6]:
            row = col.row()
            row.prop(self, p)
        col = row1.column()
        for p in list(prop_ref.keys())[6:]:
            row = col.row()
            row.prop(self, p)


classes = (
    SUPERTEXTURE_PT_panel,
    ADD_OT_supertexture,
    REMOVE_OT_supertexture,
    SUPERTEXTURE_PR_preferences)


def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
    bpy.types.Scene.st_parallax = bpy.props.BoolProperty(
        default=True, name="Use Parallax")


def unregister():
    from bpy.utils import unregister_class
    for cls in classes:
        unregister_class(cls)
    del bpy.types.Scene.st_parallax
